@@ -1,5 +1,5 @@
 # encoding=UTF-8
-# Copyright © 2008 Jakub Wilk <ubanus@users.sf.net>
+# Copyright © 2008, 2009, 2010 Jakub Wilk <ubanus@users.sf.net>
 #
 # This package is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -54,6 +54,13 @@ dict(
     ocrx_word = const.TEXT_ZONE_WORD
 ).get
 
+cuneiform_tag_to_djvu = \
+dict(
+    body = const.TEXT_ZONE_PAGE,
+    p = const.TEXT_ZONE_PARAGRAPH,
+    span = const.TEXT_ZONE_CHARACTER,
+).get
+
 djvu_zone_to_hocr = {
     const.TEXT_ZONE_PAGE: ('div', 'ocr_page'),
     const.TEXT_ZONE_COLUMN: ('div', 'ocr_carea'),
@@ -158,10 +165,98 @@ def word_break_iterator(text, locale=None):
     break_iterator.setText(text)
     return break_iterator
 
-def _replace_text(djvu_class, title, text, details, uax29):
+class Space(object):
+    pass
+
+class Zone(object):
+
+    def __init__(self, type, bbox=None, children=()):
+        self.type = type
+        self.bbox = bbox
+        self.children = list(children)
+
+    def set_bbox(self, bbox):
+        if bbox is None:
+            self._bbox = None
+        else:
+            self._bbox = tuple(bbox)
+
+    def get_bbox(self):
+        return self._bbox 
+
+    bbox = property(get_bbox, set_bbox)
+
+    @property
+    def sexpr(self):
+        return sexpr.Expression(
+            [self.type] +
+            list(self.bbox) + [
+                child.sexpr if isinstance(child, Zone) else child
+                for child in self.children
+                if not isinstance(child, Space)
+            ]
+        )
+
+    def __iter__(self):
+        return iter(self.children)
+
+    def __iadd__(self, new_children):
+        self.children += new_children
+        return self
+
+    def __getitem__(self, n):
+        return self.children[n]
+
+    def __setitem__(self, n, value):
+        self.children[n] = value
+
+    def __len__(self):
+        return len(self.children)
+
+    def __repr__(self):
+        return '%(name)s(type=%(type)r, bbox=%(bbox)r, children=%(children)r)' % dict(
+            name=type(self).__name__,
+            type=self.type,
+            bbox=self.bbox,
+            children=self.children,
+        )
+
+def _replace_cuneiform_paragraph(paragraph, settings):
+    text = ''.join(
+        ' ' if isinstance(character, Space) else character[0]
+        for character in paragraph
+    )
+    if settings.details >= const.TEXT_ZONE_LINE:
+        return [text]
+    if len(text) != len(paragraph):
+        raise ValueError
+    break_iterator = word_break_iterator(text, locale=settings.uax29)
+    i = 0
+    words = Zone(type=const.TEXT_ZONE_PARAGRAPH)
+    paragraph_bbox = BBox()
+    for j in break_iterator:
+        subtext = text[i:j]
+        if subtext.isspace():
+            i = j
+            continue
+        word_bbox = BBox()
+        for k in xrange(i, j):
+            word_bbox.update(BBox(*paragraph[k].bbox))
+        paragraph_bbox.update(word_bbox)
+        last_word = Zone(type=const.TEXT_ZONE_WORD, bbox=word_bbox)
+        words += last_word,
+        if settings.details > const.TEXT_ZONE_CHARACTER:
+            last_word += subtext,
+        else:
+            last_word += paragraph[i:j]
+        i = j
+    words.bbox = paragraph_bbox
+    return words
+
+def _replace_text(djvu_class, title, text, settings):
     if djvu_class <= const.TEXT_ZONE_LINE:
         text = text.rstrip('\n')
-    if details >= djvu_class:
+    if settings.details >= djvu_class:
         return text,
     m = BBOXES_RE.search(title)
     if not m:
@@ -180,7 +275,7 @@ def _replace_text(djvu_class, title, text, details, uax29):
     if djvu_class > const.TEXT_ZONE_WORD:
         # Split words
         words = []
-        break_iterator = word_break_iterator(text, locale=uax29)
+        break_iterator = word_break_iterator(text, locale=settings.uax29)
         i = 0
         for j in break_iterator:
             subtext = text[i:j]
@@ -190,13 +285,13 @@ def _replace_text(djvu_class, title, text, details, uax29):
             bbox = BBox()
             for k in xrange(i, j):
                 bbox.update(BBox(*coordinates[k]))
-            last_word = [const.TEXT_ZONE_WORD] + list(bbox)
+            last_word = Zone(type=const.TEXT_ZONE_WORD, bbox=bbox)
             words += last_word,
-            if details > const.TEXT_ZONE_CHARACTER:
+            if settings.details > const.TEXT_ZONE_CHARACTER:
                 last_word += subtext,
             else:
                 last_word += [
-                    [const.TEXT_ZONE_CHARACTER, x0, y0, x1, y1, ch]
+                    Zone(type=const.TEXT_ZONE_CHARACTER, bbox=(x0, y0, x1, y1), children=[ch])
                     for k in xrange(i, j)
                     for (x0, y0, x1, y1), ch in [(coordinates[k], text[k])]
                 ]
@@ -205,53 +300,58 @@ def _replace_text(djvu_class, title, text, details, uax29):
     else:
         # Split characters
         return [
-            (const.TEXT_ZONE_CHARACTER, x0, y0, x1, y1, ch)
+            Zone(type=const.TEXT_ZONE_CHARACTER, bbox=(x0, y0, x1, y1), children=[ch])
             for (x0, y0, x1, y1), ch in zip(coordinates, text)
         ]
     return text,
 
 def _rotate(obj, rotation, xform=None):
     if xform is None:
-        assert len(obj) >= 5
-        assert obj[0] == const.TEXT_ZONE_PAGE
-        assert obj[1] == obj[2] == 0
-        page_width, page_height = page_size = (obj[3], obj[4])
+        assert isinstance(obj, Zone)
+        assert obj.type == const.TEXT_ZONE_PAGE
+        assert obj.bbox[:2] == (0, 0)
+        page_size = obj.bbox[2:]
         if (rotation // 90) & 1:
-             page_width, page_height = page_size
-             xform = decode.AffineTransform((0, 0, page_height, page_width), (0, 0) + page_size)
+            xform = decode.AffineTransform((0, 0) + tuple(reversed(page_size)), (0, 0) + page_size)
         else:
             xform = decode.AffineTransform((0, 0) + page_size, (0, 0) + page_size)
         xform.mirror_y()
         xform.rotate(rotation)
-    x0, y0, x1, y1 = obj[1:5]
+    x0, y0, x1, y1 = obj.bbox
     x0, y0 = xform.inverse((x0, y0))
     x1, y1 = xform.inverse((x1, y1))
-    obj[1:5] = x0, y0, x1, y1
+    if x0 > x1:
+        x0, x1 = x1, x0
+    if y0 > y1:
+        y0, y1 = y1, y0
+    obj.bbox = x0, y0, x1, y1
     for child in obj:
-        if isinstance(child, list):
+        if isinstance(child, Zone):
             _rotate(child, rotation, xform)
-        elif isinstance(child, (sexpr.Symbol, int, basestring)):
-            pass
-        else:
-            raise TypeError('%r in not in: list, int, basestring, Symbol' % type(child).__name__)
 
-def _scan(node, buffer, parent_bbox, page_size=None, details=TEXT_DETAILS_WORD, uax29=None):
+def _scan(node, buffer, parent_bbox, settings):
     def look_down(buffer, parent_bbox):
         for child in node.iterchildren():
-            _scan(child, buffer, parent_bbox, page_size, details, uax29)
-            if node.tail and node.tail.strip():
-                buffer.append(node.tail)
+            _scan(child, buffer, parent_bbox, settings)
+            if child.tail:
+                if child.tail.strip():
+                    buffer += child.tail,
+                else:
+                    buffer += Space(),
         if node.text and node.text.strip():
-            buffer.append(node.text)
+            buffer += node.text,
     if not isinstance(node.tag, basestring):
         return
-    hocr_classes = (node.get('class') or '').split()
-    djvu_class = None
-    for hocr_class in hocr_classes:
-        djvu_class = hocr_class_to_djvu(hocr_class)
-        if djvu_class:
-            break
+    if settings.cuneiform:
+        djvu_class = cuneiform_tag_to_djvu(node.tag)
     else:
+        hocr_classes = (node.get('class') or '').split()
+        djvu_class = None
+        for hocr_class in hocr_classes:
+            djvu_class = hocr_class_to_djvu(hocr_class)
+            if djvu_class:
+                break
+    if not djvu_class:
         look_down(buffer, parent_bbox)
         return
     title = node.get('title') or ''
@@ -263,53 +363,70 @@ def _scan(node, buffer, parent_bbox, page_size=None, details=TEXT_DETAILS_WORD,
         parent_bbox.update(bbox)
     if djvu_class is const.TEXT_ZONE_PAGE:
         if not bbox:
-            m = IMAGE_RE.search(title)
-            if m is None:
-                raise errors.MalformedHocr("cannot determine page's bbox")
-            page_size = image_size.get_image_size(m.group('file_name'))
-            page_width, page_height = page_size
+            if settings.page_size is None:
+                m = IMAGE_RE.search(title)
+                if m is None:
+                    raise errors.MalformedHocr("cannot determine page's bbox")
+                settings.page_size = image_size.get_image_size(m.group('file_name'))
+            page_width, page_height = settings.page_size
             bbox = BBox(0, 0, page_width - 1, page_height - 1)
             parent_bbox.update(bbox)
         else:
             if (bbox.x0, bbox.y0) != (0, 0):
                 raise errors.MalformedHocr("page's bbox should start with (0, 0)")
-            page_size = bbox.x1, bbox.y1
-    result = [sexpr.Symbol(djvu_class)]
-    result += [None] * 4
+            settings.page_size = bbox.x1, bbox.y1
+    result = Zone(type=djvu_class)
     look_down(result, bbox)
-    if isinstance(result[-1], basestring):
-        result[-1:] = _replace_text(djvu_class, title, result[-1], details, uax29)
+    if len(result) > 0 and isinstance(result[-1], basestring):
+        result[-1:] = _replace_text(djvu_class, title, result[-1], settings)
+    elif settings.cuneiform and djvu_class is const.TEXT_ZONE_PARAGRAPH:
+        result[:] = _replace_cuneiform_paragraph(result[:], settings)
     if not bbox and not len(node):
         return
-    if page_size is None:
+    if settings.page_size is None:
         raise errors.MalformedHocr('unable to determine page size')
-    result[1], result[2], result[3], result[4] = bbox
-    if len(result) == 5:
-        result.append('')
-    buffer.append(result)
+    result.bbox = bbox
+    if len(result) == 0:
+        result += '',
+    buffer += result,
 
-def scan(node, rotation=0, details=TEXT_DETAILS_WORD, uax29=None):
+def scan(node, settings):
     buffer = []
-    _scan(node, buffer, BBox(), details=details, uax29=uax29)
-    for obj in buffer:
-        _rotate(obj, rotation)
+    _scan(node, buffer, BBox(), settings)
+    # Buffer may contain also superfluous Space objects. Let's strip them.
+    buffer = [zone for zone in buffer if isinstance(zone, Zone)]
+    for zone in buffer:
+        _rotate(zone, settings.rotation)
     return buffer
 
-def extract_text(stream, rotation=0, details=TEXT_DETAILS_WORD, uax29=None):
+class ExtractSettings(object):
+
+    def __init__(self, rotation=0, details=TEXT_DETAILS_WORD, uax29=None, page_size=None, cuneiform=False):
+        self.rotation = rotation
+        self.details = details
+        if uax29 is not None:
+            icu = get_icu()
+            if uax29 is True:
+                uax29 = icu.Locale()
+            else:
+                uax29 = icu.Locale(uax29)
+        self.uax29 = uax29
+        self.page_size = page_size
+        self.cuneiform = cuneiform
+
+def extract_text(stream, **kwargs):
     '''
     Extract DjVu text from an hOCR stream.
 
     details: TEXT_DETAILS_LINES or TEXT_DETAILS_WORD or TEXT_DETAILS_CHAR
     uax29: None or a PyICU locale
     '''
-    if uax29 is not None:
-        icu = get_icu()
-        if uax29 is True:
-            uax29 = icu.Locale()
-        else:
-            uax29 = icu.Locale(uax29)
+    settings = ExtractSettings(**kwargs)
     doc = ET.parse(stream, ET.HTMLParser())
-    scan_result = scan(doc.find('/body'), rotation=rotation, details=details, uax29=uax29)
-    return sexpr.Expression(scan_result)
+    if doc.find('/head/meta[@name="ocr-capabilities"]') is None:
+        # Wild guess:
+        settings.cuneiform = True
+    scan_result = scan(doc.find('/body'), settings)
+    return [zone.sexpr for zone in scan_result]
 
 # vim:ts=4 sw=4 et
